#include "../Framework/Framework.h"

#include "Poly.h"

class MainApp : public OpenGLApp {
protected:
	TextureID tex, texBump, lightTexture, vectorField, distanceMap;

	Set <Poly> polys;
	Vertex light;

public:
	bool load();
	bool unload();
	bool drawFrame();
	
	void resetCamera();
	void selectPixelFormat(PixelFormat &pixelFormat);
};

void MainApp::resetCamera(){
	position = Vertex(500,-220,-500);
	Wx = 0;
	Wy = PI / 4;
	Wz = 0;
}

void MainApp::selectPixelFormat(PixelFormat &pixelFormat){
	pixelFormat.redBits   = 8;
	pixelFormat.greenBits = 8;
	pixelFormat.blueBits  = 8;
	pixelFormat.alphaBits = 8;

	pixelFormat.depthBits = 16;
}

bool MainApp::load(){
	String supported, missing;

    if (GL_ARB_texture_env_dot3)    supported += "GL_ARB_texture_env_dot3\n";    else missing += "GL_ARB_texture_env_dot3\n";
    if (GL_EXT_texture3D)           supported += "GL_EXT_texture3D\n";           else missing += "GL_EXT_texture3D\n";
    if (GL_EXT_texture_edge_clamp)  supported += "GL_EXT_texture_edge_clamp\n";  else missing += "GL_EXT_texture_edge_clamp\n";
    if (GL_EXT_blend_subtract)      supported += "GL_EXT_blend_subtract\n";      else missing += "GL_EXT_blend_subtract\n";
	
	int nTmu = renderer->getSimultaneousTextures();
	if (nTmu < 4) missing += "4 Texture units\n";

    if (missing.getLength() > 0){
        ALERT("The following required OpenGL extensions are supported:\n\n" + supported +
				   "\nToo bad, these required OpenGL extensions aren't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.", "Missing extensions");
        return false;
    }


	vectorField = renderer->addVectorField(64, TEX_NOMIPMAPPING);
	distanceMap = renderer->addDistanceMap(64, TEX_NOMIPMAPPING);

	tex      = renderer->addTexture("../Textures/16tile01.jpg");
	texBump  = renderer->addTexture("../Textures/16tile01Bump.png", TEX_NORMALMAP | TEX_USE_16BIT);

	textTexture  = renderer->addTexture("../Textures/font2.dds");

	lightTexture = renderer->addParticleTexture(64);

	projection.loadProjectionMatrix(90 * PI / 180, 0.75f, 0.03f, 2000);

	polys.add(Poly(Vertex(-512,  512,  512), Vertex( 512,  512,  512), Vertex(-512, -256,  512), tex, texBump, 2, 2));
	polys.add(Poly(Vertex(-512,  512, -512), Vertex( 512,  512, -512), Vertex(-512,  512,  512), tex, texBump, 2, 2));
	polys.add(Poly(Vertex( 512,  512,  512), Vertex( 512,  512, -512), Vertex( 512, -256,  512), tex, texBump, 2, 2));
	polys.add(Poly(Vertex( 512,  512, -512), Vertex(-512,  512, -512), Vertex( 512, -256, -512), tex, texBump, 2, 2));
	polys.add(Poly(Vertex(-512,  512, -512), Vertex(-512,  512,  512), Vertex(-512, -256, -512), tex, texBump, 2, 2));

	polys.add(Poly(Vertex(-512, -256,  512), Vertex( 512, -256,  512), Vertex(-512, -256, -512), tex, texBump, 2, 2));
	
	polys.add(Poly(Vertex(-120,    0,  120), Vertex( 120,    0,  120), Vertex(-120,    0, -120), tex, texBump, 1, 1));
	// Ugh, ugly hack instead of adding real dubblesided lighting 
	polys.add(Poly(Vertex(-120,   -2, -120), Vertex( 120,   -2, -120), Vertex(-120,   -2,  120), tex, texBump, 1, 1));

	return true;
}

bool MainApp::unload(){
	polys.clear();

	return true;
}

bool MainApp::drawFrame(){

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projection);

	modelView.loadRotateZXY(-Wx, -Wy, -Wz);
	modelView.translate(-position);

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelView);

	glClear(GL_DEPTH_BUFFER_BIT);

	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

	light = Vertex(15 * sinf(currTime * 2.5f), 15 * sinf(currTime * 3.24f) + 30, 35 * cosf(currTime * 0.73f));
	light.normalize();
	light *= 70 * (sinf(currTime + 0.43f) + 3.0f);
	

	for (unsigned int i = 0; i < polys.getSize(); i++){
		polys[i].drawLit(renderer, vectorField, distanceMap, position, light);
	}
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	renderer->setTextures(distanceMap);
	renderer->apply();

	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_TRUE);

	glCullFace(GL_FRONT);
	polys[6].drawShadowVolume(position, light);

	glCullFace(GL_BACK);
	renderer->setTextures(distanceMap);
	renderer->setBlending(GL_ONE, GL_ONE, GL_FUNC_REVERSE_SUBTRACT_EXT);
	renderer->apply();

	polys[6].drawShadowVolume(position, light);

	if (polys[6].coversLight(position, light)){
		glDisable(GL_DEPTH_TEST);
		glCullFace(GL_FRONT);
	} else {
		glCullFace(GL_BACK);
	}

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	renderer->setBlending(GL_ZERO, GL_ONE_MINUS_DST_ALPHA);
	renderer->apply();
	polys[6].drawShadowVolume(position, light);

	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	renderer->setBlending(GL_ONE, GL_ONE);
	renderer->setTextures(lightTexture);
	renderer->apply();

	glDepthMask(GL_FALSE);

	glColor3f(1.0f,0.4f,0.1f);
	Vertex dx(modelView[0], modelView[4], modelView[8]);
	Vertex dy(modelView[1], modelView[5], modelView[9]);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(light - 30 * dx - 30 * dy);
		
		glTexCoord2f(1,0);
		glVertex3fv(light + 30 * dx - 30 * dy);

		glTexCoord2f(1,1);
		glVertex3fv(light + 30 * dx + 30 * dy);

		glTexCoord2f(0,1);
		glVertex3fv(light - 30 * dx + 30 * dy);
	glEnd();

	glDepthMask(GL_TRUE);

	return true;
}

DECLARE_MAIN_APP(MainApp)
